﻿using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Chapter1
{
    namespace Sample01
    {
        // Mutable Address structure.
        public struct Address
        {
            private string state;
            private int zipCode;

            // Rely on the default system-generated
            // constructor.

            public string Line1
            {
                get;
                set;
            }
            public string Line2
            {
                get;
                set;
            }
            public string City
            {
                get;
                set;
            }
            public string State
            {
                get => state;
                set
                {
                    ValidateState(value);
                    state = value;
                }
            }

            public int ZipCode
            {
                get => zipCode;
                set
                {
                    ValidateZip(value);
                    zipCode = value;
                }
            }

            // other details omitted.


            private void ValidateState(String value)
            {
                throw new NotImplementedException();
            }
            private void ValidateZip(Int32 value)
            {
                throw new NotImplementedException();
            }
        }

        public static class Test
        {
            public static void Tests()
            {
                // Example usage:
                Address a1 = new Address();
                a1.Line1 = "111 S. Main";
                a1.City = "Anytown";
                a1.State = "IL";
                a1.ZipCode = 61111;
                // Modify:
                a1.City = "Ann Arbor"; // Zip, State invalid now.
                a1.ZipCode = 48103; // State still invalid now.
                a1.State = "MI"; // Now fine.
            }
        }
    }

    namespace Sample2
    {
        public struct Address
        {
            // remaining details elided
            public string Line1 { get; }
            public string Line2 { get; }
            public string City { get; }
            public string State { get; }
            public int ZipCode { get; }

            public Address(string line1,
                string line2,
                string city,
                string state,
                int zipCode) :
                this()
            {
                Line1 = line1;
                Line2 = line2;
                City = city;
                ValidateState(state);
                State = state;
                ValidateZip(zipCode);
                ZipCode = zipCode;
            }

            private void ValidateZip(Int32 zipCode)
            {
                throw new NotImplementedException();
            }

            private void ValidateState(String state)
            {
                throw new NotImplementedException();
            }
        }

        public static class Test
        {
            public static void Tests()
            {
                // Create an address:
                Address a2 = new Address("111 S. Main",
                  "", "Anytown", "IL", 61111);

                // To change, re-initialize:
                a2 = new Address(a2.Line1,
                    a2.Line2, "Ann Arbor", "MI", 48103);
            }
        }
    }

    namespace Sample3
    {
        public class Phone
        {
            internal static Phone GeneratePhoneNumber()
            {
                throw new NotImplementedException();
            }
        }

        // Almost immutable: there are holes that would
        // allow state changes.
        public struct PhoneList
        {
            private readonly Phone[] phones;

            public PhoneList(Phone[] ph)
            {
                phones = ph;
            }

            // Use Select() to prevent cast back to array
            // See Item 36 in Effective C#
            public IEnumerable<Phone> Phones => phones.Select(p => p);
        }

        public static class Test
        {
            public static void Tests()
            {
                Phone[] phones = new Phone[10];
                // initialize phones
                PhoneList pl = new PhoneList(phones);

                // Modify the phone list:
                // also modifies the internals of the (supposedly)
                // immutable object.
                phones[5] = Phone.GeneratePhoneNumber();
            }
        }
    }

    namespace Sample4
    {
        public class Phone
        {
            internal static Phone GeneratePhoneNumber()
            {
                throw new NotImplementedException();
            }
        }
        // Immutable: A copy is made at construction.
        public struct PhoneList
        {
            private readonly ImmutableList<Phone> phones;

            public PhoneList(Phone[] ph)
            {
                phones = ph.ToImmutableList();
            }

            public IEnumerable<Phone> Phones => phones;
        }

        public static class Test
        {
            public static void Tests()
            {
                Phone[] phones = new Phone[10];
                // initialize phones
                PhoneList p2 = new PhoneList(phones);

                // Modify the phone list:
                // Does not modify the copy in pl.
                phones[5] = Phone.GeneratePhoneNumber();

            }
        }
    }

    class Item03
    {

        public static void Tests()
        {

        }
    }
}
