﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace Chapter3_Task
{
    class Item31
    {
        public class ProgressReporter : IProgress<(int percent, string message)>
        {
            public void Report((int percent, string message) value)
            {
                Console.WriteLine($"{value.percent} completed: {value.message}");
            }
        }
        public static void Tests()
        {
            var generator = new Item31();

            generator.RunPayroll(DateTime.Now, new ProgressReporter());

            var cts = new CancellationTokenSource();
            generator.RunPayroll(DateTime.Now, cts.Token);
            // to cancel:
            cts.Cancel();
        }

        public async Task RunPayroll2(DateTime payrollPeriod)
        {
            // Step 1: Calculate hours and pay
            var payrollData = await RetieveEmployeePayrollDataFor(
                payrollPeriod);

            // Step 2: Calculate tax
            var taxReporting = new Dictionary<EmployeePayrollData,
                TaxWithholding>();
            foreach(var employee in payrollData)
            {
                var taxWitholding = await RetrieveTaxData(employee);
                taxReporting.Add(employee, taxWitholding);
            }

            // Step 3: generate and email paystub documents
            var paystubs = new List<Task>();
            foreach(var payrollItem in taxReporting)
            {
                var payrollTask = GeneratePayrollDocument(
                    payrollItem.Key, payrollItem.Value);
                var emailTask = payrollTask.ContinueWith(
                    paystub => EmailPaystub(payrollItem.Key.Email, paystub.Result));
                paystubs.Add(emailTask);
            }
            await Task.WhenAll(paystubs);

            // Step 4: Deposit pay
            var depositTasks = new List<Task>();
            foreach(var payrollItem in taxReporting)
            {
                depositTasks.Add(MakeDeposit(payrollItem.Key, payrollItem.Value));
            }
            await Task.WhenAll(depositTasks);
            // Step 5: Close payroll period
            await ClosePayrollPeriod(payrollPeriod);
        }

        public async Task RunPayroll2(DateTime payrollPeriod, 
            CancellationToken cancellationToken)
        {
            // Step 1: Calculate hours and pay
            var payrollData = await RetieveEmployeePayrollDataFor(
                payrollPeriod);
            cancellationToken.ThrowIfCancellationRequested();

            // Step 2: Calculate tax
            var taxReporting = new Dictionary<EmployeePayrollData,
                TaxWithholding>();
            foreach (var employee in payrollData)
            {
                var taxWitholding = await RetrieveTaxData(employee);
                taxReporting.Add(employee, taxWitholding);
            }
            cancellationToken.ThrowIfCancellationRequested();

            // Step 3: generate and email paystub documents
            var paystubs = new List<Task>();
            foreach (var payrollItem in taxReporting)
            {
                var payrollTask = GeneratePayrollDocument(
                    payrollItem.Key, payrollItem.Value);
                var emailTask = payrollTask.ContinueWith(
                    paystub => EmailPaystub(payrollItem.Key.Email, paystub.Result));
                paystubs.Add(emailTask);
            }
            await Task.WhenAll(paystubs);
            cancellationToken.ThrowIfCancellationRequested();

            // Step 4: Deposit pay
            var depositTasks = new List<Task>();
            foreach (var payrollItem in taxReporting)
            {
                depositTasks.Add(MakeDeposit(payrollItem.Key, payrollItem.Value));
            }
            await Task.WhenAll(depositTasks);

            // Step 5: Close payroll period
            await ClosePayrollPeriod(payrollPeriod);
        }

        public async Task RunPayroll2(DateTime payrollPeriod, 
            IProgress<(int, string)> progress)
        {
            progress?.Report((0, "Starting Payroll"));
            // Step 1: Calculate hours and pay
            var payrollData = await RetieveEmployeePayrollDataFor(
                payrollPeriod);

            progress?.Report((20, "Retrieved employees and hours"));

            // Step 2: Calculate tax
            var taxReporting = new Dictionary<EmployeePayrollData,
                TaxWithholding>();
            foreach (var employee in payrollData)
            {
                var taxWitholding = await RetrieveTaxData(employee);
                taxReporting.Add(employee, taxWitholding);
            }
            progress?.Report((40, "Calculated Witholding"));

            // Step 3: generate and email paystub documents
            var paystubs = new List<Task>();
            foreach (var payrollItem in taxReporting)
            {
                var payrollTask = GeneratePayrollDocument(
                    payrollItem.Key, payrollItem.Value);
                var emailTask = payrollTask.ContinueWith(
                    paystub => EmailPaystub(payrollItem.Key.Email, paystub.Result));
                paystubs.Add(emailTask);
            }
            await Task.WhenAll(paystubs);
            progress?.Report((60, "Emailed Paystubs"));

            // Step 4: Deposit pay
            var depositTasks = new List<Task>();
            foreach (var payrollItem in taxReporting)
            {
                depositTasks.Add(MakeDeposit(payrollItem.Key, payrollItem.Value));
            }
            await Task.WhenAll(depositTasks);
            progress?.Report((80, "Deposited pay"));

            // Step 5: Close payroll period
            await ClosePayrollPeriod(payrollPeriod);
            progress?.Report((100, "complete"));
        }

        public Task RunPayroll(DateTime payrollPeriod) =>
            RunPayroll(payrollPeriod, new CancellationToken(), null);

        public Task RunPayroll(DateTime payrollPeriod,
            CancellationToken cancellationToken) =>
            RunPayroll(payrollPeriod, cancellationToken, null);

        public Task RunPayroll(DateTime payrollPeriod,
            IProgress<(int, string)> progress) =>
            RunPayroll(payrollPeriod, new CancellationToken(), progress);

        public async Task RunPayroll(DateTime payrollPeriod,
            CancellationToken cancelationToken,
            IProgress<(int, string)> progress)
        {
            progress?.Report((0, "Starting Payroll"));
            // Step 1: Calculate hours and pay
            var payrollData = await RetieveEmployeePayrollDataFor(
                payrollPeriod);
            cancelationToken.ThrowIfCancellationRequested();
            progress?.Report((20, "Retrieved employees and hours"));

            // Step 2: Calculate tax
            var taxReporting = new Dictionary<EmployeePayrollData,
                TaxWithholding>();
            foreach (var employee in payrollData)
            {
                var taxWitholding = await RetrieveTaxData(employee);
                taxReporting.Add(employee, taxWitholding);
            }
            cancelationToken.ThrowIfCancellationRequested();
            progress?.Report((40, "Calculated Witholding"));

            // Step 3: generate and email paystub documents
            var paystubs = new List<Task>();
            foreach (var payrollItem in taxReporting)
            {
                var payrollTask = GeneratePayrollDocument(
                    payrollItem.Key, payrollItem.Value);
                var emailTask = payrollTask.ContinueWith(
                    paystub => EmailPaystub(payrollItem.Key.Email, paystub.Result));
                paystubs.Add(emailTask);
            }
            await Task.WhenAll(paystubs);
            cancelationToken.ThrowIfCancellationRequested();
            progress?.Report((60, "Emailed Paystubs"));

            // Step 4: Deposit pay
            var depositTasks = new List<Task>();
            foreach (var payrollItem in taxReporting)
            {
                depositTasks.Add(MakeDeposit(payrollItem.Key, payrollItem.Value));
            }
            await Task.WhenAll(depositTasks);
            progress?.Report((80, "Deposited pay"));

            // Step 5: Close payroll period
            await ClosePayrollPeriod(payrollPeriod);
            cancelationToken.ThrowIfCancellationRequested();
            progress?.Report((100, "complete"));
        }

        private Task ClosePayrollPeriod(DateTime payrollPeriod)
        {
            throw new NotImplementedException();
        }

        private Task MakeDeposit(EmployeePayrollData key, TaxWithholding value)
        {
            throw new NotImplementedException();
        }

        private Task EmailPaystub(object email, PdfDocument paystub)
        {
            throw new NotImplementedException();
        }

        private Task<PdfDocument> GeneratePayrollDocument(EmployeePayrollData key, TaxWithholding value)
        {
            throw new NotImplementedException();
        }

        private Task<TaxWithholding> RetrieveTaxData(EmployeePayrollData employee)
        {
            throw new NotImplementedException();
        }

        private Task<IEnumerable<EmployeePayrollData>> RetieveEmployeePayrollDataFor(DateTime payrollPeriod)
        {
            throw new NotImplementedException();
        }
    }

    public class EmployeePayrollData
    {
        public int Id;
        public string Name;
        public decimal Hours;

        public string Email;
    }
    public class TaxWithholding
    {
        public decimal FederalWithholding;
        public decimal StateWithholding;
    }
}
